/+
 ' Program         : randomdata.exe (Console App)
 '                 : Random personnel names and addresses generator.
 ' Source          : randomdata.d v0.25 Beta
 ' Author          : David L. 'SpottedTiger' Davis
 ' Source Platform : Intel x86 / WinXP SP2 
 ' Target Platform : Should work on any CPU / OS combo that D compiles on.
 '                 :
 ' Complier        : DigitalMars "D" aka Mars v0.107
 ' Created Date    : 29.Nov.04 Compiled and tested with dmd v0.107
 ' Modified Date   : 01.Dec.04 Recompiled and tested with dmd v0.108
 '                 : 29.Jan.05 Recompiled and tested with dmd v0.111
 '                 : 10.Jun.05 Recompiled and tested with dmd v0.126
 '                 : 17.Sep.05 Recompiled and tested with dmd v0.131
 '                 : 18.Sep.05 Made some performance improvements to
 '                 :           the code, which saved five minutes using 
 '                 :           the default configuration parameters to 
 '                 :           create 10,000 entries. Also added the
 '                 :           setPosTagMarkers() function.
 '                 : 26.Oct.05 Recompiled and Tested with dmd v0.137.
 '                 : 14.Jan.06 Recompiled and Tested with dmd v0.143.
 '                 : 26.Mar.06 Recompiled and Tested with dmd v0.150, and
 '                 :           added the genBirthDate() function which takes
 '                 :           in the age of the person as an input parameter
 '                 :           to create a birth date for them.
 '                 : 28.May.06 Recompiled and Tested with dmd v0.159.
 '                 : 19.Jul.06 All "Imports" are now private by default  
 '                 :           (which shouldn't be a problem, since I  
 '                 :           always use the private keyword) in 
 '                 :           dmd v0.163, recompiled and tested.
 '                 : 26.Jul.06 Updated the getRandomNumber() function.  
 '                 : 11.Dec.06 Recompiled and Tested with dmd v0.177.
 '                 : 05.Jan.0& Recompiled and Tested with dmd v1.0.
 '                 :
 ' Requirements    : std.string, std.conv, std.stdio, std.random, std.date, std.file,
 '                 : toyxml.d v0.24, stringplus.d v0.13, stringbuffer.d v0.12, and
 '                 : dateutils.d v0.3 (the last four are included in the zipfile)
 '                 :
 ' Licence         : Public Domain
 ' ---------------------------------------------------------------------------
 ' Note: That local copies of toyxml.d, stringplus.d, stringbuffer.d, and dateutils.d 
 '       will be placed in the randomdata.d file in "\dmd\randomdata" directory so that 
 '       everything is together...also all the *.xml files are to be in the "xml" sub-dirertory.
 '
 ' Public functions:
 ' -----------------
 ' 01) int getStateFilenames( in char[], inout udtXmlInfo[] )
 ' 02) int getNameFilenames( in char[], inout udtXmlInfo[] )
 ' 03) int getAddrFilenames( in char[], inout udtXmlInfo[] )
 ' 04) void loadStateXMLs( in char[], inout udtXmlInfo[], inout toyXML[] )
 ' 05) void loadNameXMLs( in char[], inout udtXmlInfo[], inout toyXML[] ) 
 ' 06) void loadAddrXMLs( in char[], inout udtXmlInfo[], inout toyXML[] )
 ' 07) void genFemaleName( in udtXmlInfo[], in toyXML[], out char[], out char[], out char[] ) 
 ' 08) void genMaleName( in udtXmlInfo[], in toyXML[], out char[], out char[], out char[] ) 
 ' 09) void genAddress( in udtXmlInfo[], in toyXML[], out char[], out char[], out char[] )
 ' 10) void genCityStateZipCode( in udtXmlInfo[], in toyXML[], out char[], out char[], out char[] out char[] )
 ' 11) void getOutputOptions( in char[], out char[], out uint )
 ' 12) uint getRandomNumber( in uint = 0x0, in uint = 0xFFFFFFFF )
 ' 13) bool setPosTagMarkers( in int, in char[], in udtXmlInfo[], in toyXML[] )
 ' 14) char[] genBirthDate( in uint )
 '
 ' =====================================
 ' To compile the project:
 ' C:\dmd\randomdata>..\bin\dmd randomdata.d toyxml.d stringplus.d stringbuffer.d dateutils.d
 '
 ' To run after compiling:
 ' C:\dmd\randomdata>randomdata 
 +/
module randomdata;

private import std.stdio;
private import std.string;
private import std.file;
private import std.conv;
private import stringplus;
private import toyxml;
private import stringbuffer;
private import dateutils;

// for getRandomNumber()
private import std.random;
private import std.date;

struct udtXmlInfo
{
    char[] sXMLFile   = ""c;
    char[] sRootTag   = ""c;
    int    iNumOfRows = 0;
    uint[] uiMarkers;
}

enum
{
    FEMALEFIRST  = 0,
    FEMALEMIDDLE = 1,
    MALEFIRST    = 2,
    MALEMIDDLE   = 3,
    LASTNAMES    = 4,
    BOTHMIDDLE   = 5,
    DESIGNATOR   = 0,
    SUFFIX       = 1,
    STREETNAME   = 2,
    FEMALE       = 0,
    MALE         = 1
}

// Loads the State XML Filenames into an array.
int getStateFilenames( in char[] sXMLConfig, inout udtXmlInfo[] arrStates )
{
    toyXML oXML         = new toyXML;
    int    iNumOfStates = 0;
    char[] sFileExt     = ""c;
    char[][] sarrStates;

    //example: sXMLConfig = r"C:\dmd\randomdata\randomdata.xml";
    oXML.loadXMLDocument(sXMLConfig);

    sarrStates = std.string.split(oXML.getNodeValue("/config/usa/states_selected" ), ",");

    if (sarrStates.length == 0)
        return 0;

    sFileExt = oXML.getNodeValue("/config/usa/zipcode_fileext"c);
    arrStates.length = sarrStates.length;

    for (uint uix = 0; uix < sarrStates.length; uix++)
    {
        // example: "AK" ~ "_zipcodes.xml"
        arrStates[uix].sXMLFile = sarrStates[uix] ~ sFileExt;

        // Won't know this until the file(s) is read in.
        arrStates[uix].iNumOfRows = 0;   
    }

    delete oXML;
    
    return iNumOfStates;

} // end int getStateFilenames( in char[], inout udtStates[] )

// Loads the Name XML Filenames into an array.
int getNameFilenames(in char[] sXMLConfig, inout udtXmlInfo[] arrNames)
{
    toyXML oXML = new toyXML;

    //example: sXMLConfig = r"C:\dmd\randomdata\randomdata.xml";
    oXML.loadXMLDocument(sXMLConfig);

    arrNames.length = 6;

    arrNames[FEMALEFIRST].sXMLFile = oXML.getNodeValue("/config/usa/female_first_names"c);
    arrNames[FEMALEFIRST].sRootTag = oXML.getAttrValue("/config/usa/female_first_names"c, "roottag"c); 
    //writefln( "roottag=%s", arrNames[FEMALEFIRST].sRootTag );
    arrNames[FEMALEFIRST].iNumOfRows = 0;

    arrNames[FEMALEMIDDLE].sXMLFile = oXML.getNodeValue("/config/usa/female_middle_names"c);
    arrNames[FEMALEMIDDLE].sRootTag = oXML.getAttrValue("/config/usa/female_middle_names"c, "roottag"c); 
    //writefln( "roottag=%s", arrNames[FEMALEMIDDLE].sRootTag );
    arrNames[FEMALEMIDDLE].iNumOfRows = 0;

    arrNames[MALEFIRST].sXMLFile = oXML.getNodeValue("/config/usa/male_first_names"c);
    arrNames[MALEFIRST].sRootTag = oXML.getAttrValue("/config/usa/male_first_names"c, "roottag"c); 
    //writefln( "roottag=%s", arrNames[MALEFIRST].sRootTag );
    arrNames[MALEFIRST].iNumOfRows = 0;

    arrNames[MALEMIDDLE].sXMLFile = oXML.getNodeValue("/config/usa/male_middle_names"c);
    arrNames[MALEMIDDLE].sRootTag = oXML.getAttrValue("/config/usa/male_middle_names"c, "roottag"c); 
    //writefln( "roottag=%s", arrNames[MALEMIDDLE].sRootTag );
    arrNames[MALEMIDDLE].iNumOfRows = 0;

    arrNames[LASTNAMES].sXMLFile = oXML.getNodeValue("/config/usa/last_names"c);
    arrNames[LASTNAMES].sRootTag = oXML.getAttrValue("/config/usa/last_names"c, "roottag"c); 
    //writefln( "roottag=%s", arrNames[LASTNAMES].sRootTag );
    arrNames[LASTNAMES].iNumOfRows = 0;

    arrNames[BOTHMIDDLE].sXMLFile = oXML.getNodeValue("/config/usa/bothsexes_middle_names"c);
    arrNames[BOTHMIDDLE].sRootTag = oXML.getAttrValue("/config/usa/bothsexes_middle_names"c, "roottag"c); 
    //writefln( "roottag=%s", arrNames[BOTHMIDDLE].sRootTag );
    arrNames[BOTHMIDDLE].iNumOfRows = 0;

    delete oXML;

    return 6;
    
} // end int getNameFilenames( in char[], inout udtXmlInfo[] )

// Loads the Address XML Filenames into an array.
int getAddrFilenames(in char[] sXMLConfig, inout udtXmlInfo[] arrAddrs)
{
    toyXML oXML = new toyXML;

    //example: sXMLConfig = r"C:\dmd\randomdata\randomdata.xml";
    oXML.loadXMLDocument(sXMLConfig);

    arrAddrs.length = 3;

    arrAddrs[DESIGNATOR].sXMLFile = oXML.getNodeValue("/config/usa/designators"c);
    arrAddrs[DESIGNATOR].sRootTag = oXML.getAttrValue("/config/usa/designators"c, "roottag"c); 
    //writefln( "roottag=%s", arrAddrs[DESIGNATOR].sRootTag );
    arrAddrs[DESIGNATOR].iNumOfRows = 0;

    arrAddrs[SUFFIX].sXMLFile = oXML.getNodeValue("/config/usa/suffixes"c);
    arrAddrs[SUFFIX].sRootTag = oXML.getAttrValue("/config/usa/suffixes"c, "roottag"c);
    //writefln( "roottag=%s", arrAddrs[SUFFIX].sRootTag );
    arrAddrs[SUFFIX].iNumOfRows = 0;

    arrAddrs[STREETNAME].sXMLFile = oXML.getNodeValue("/config/usa/streets"c);
    arrAddrs[STREETNAME].sRootTag = oXML.getAttrValue("/config/usa/streets"c, "roottag"c);
    //writefln( "roottag=%s", arrAddrs[STREETNAME].sRootTag );
    arrAddrs[STREETNAME].iNumOfRows = 0;

    delete oXML;

    return 3;
    
} // end int getAddrFilenames( in char[], inout udtXmlInfo[] )

void loadStateXMLs
( 
    in char[] sPath, 
    inout udtXmlInfo[] arrStates, 
    inout toyXML[] oXmlStates
)
{
    char[] sNode  = ""c;
    char[] sCount = "0"c;

    // Load States Selected to load from Config XML File.
    getStateFilenames(sPath ~ r"\randomdata.xml"c, arrStates);

    if (arrStates.length == 0)
        return ; 

    // Load the State XML Documents into memory
    oXmlStates.length = arrStates.length;

    for (uint uix = 0; uix < arrStates.length; uix++)
    {
        oXmlStates[uix] = new toyXML;
        oXmlStates[uix].loadXMLDocument(sPath ~ r"\xml\"c ~ arrStates[uix].sXMLFile);
        sNode = std.string.replace(std.string.tolower(arrStates[uix].sXMLFile), ".xml"c, ""c);
        sCount = oXmlStates[uix].getAttrValue(sNode, "count"c);
        
        arrStates[uix].iNumOfRows = std.conv.toInt((sCount == ""c ? "0"c : sCount));
        //writefln( "arrStates.sFilename=%s, iCount=%d", arrStates[ uix ].sXMLFile, arrStates[ uix ].iNumOfRows );
    }

} // end loadStateXMLs( in char[], inout udtXmlInfo[], inout toyXML[] )

void loadNameXMLs
( 
    in char[] sPath, 
    inout udtXmlInfo[] arrNames, 
    inout toyXML[] oXmlNames 
)
{
    char[] sCount = "0"c;

    // Load Names from Config XML File.
    getNameFilenames(sPath ~ r"\randomdata.xml"c, arrNames);

    // Load the Name XML Documents into memory
    oXmlNames.length = arrNames.length;
    for (uint uix = 0; uix < arrNames.length; uix++)
    {
        oXmlNames[uix] = new toyXML;
        oXmlNames[uix].loadXMLDocument(sPath ~ r"\xml\"c ~ arrNames[uix].sXMLFile);
        sCount = oXmlNames[uix].getAttrValue(arrNames[uix].sRootTag, "count"c);
        arrNames[uix].iNumOfRows = std.conv.toInt((sCount == ""c ? "0"c : sCount));
        //writefln( "arrNames.sFilename=%s, iCount=%d", arrNames[ uix ].sXMLFile, arrNames[ uix ].iNumOfRows );
    }

} // end loadNameXMLs( in char[], inout udtXmlInfo[], inout toyXML[] )

void loadAddrXMLs
( 
    in char[] sPath, 
    inout udtXmlInfo[] arrAddrs, 
    inout toyXML[] oXmlAddrs 
)
{
    char[] sCount = "0"c;

    // Load Addresses from Config XML File.
    getAddrFilenames(sPath ~ r"\randomdata.xml"c, arrAddrs);

    // Load the Addresses XML Documents into memory
    oXmlAddrs.length = arrAddrs.length;
    for (uint uix = 0; uix < arrAddrs.length; uix++)
    {
        oXmlAddrs[uix] = new toyXML;
        oXmlAddrs[uix].loadXMLDocument(sPath ~ r"\xml\"c ~ arrAddrs[uix].sXMLFile);
        sCount = oXmlAddrs[uix].getAttrValue(arrAddrs[uix].sRootTag, "count"c);

        arrAddrs[uix].iNumOfRows = std.conv.toInt((sCount == ""c ? "0"c : sCount));
        //writefln( "arrAddrs.sFilename=%s, iCount=%d", arrAddrs[ uix ].sXMLFile, arrAddrs[ uix ].iNumOfRows );
    }

} // end loadNameXMLs( in char[], inout udtXmlInfo[], inout toyXML[] )

void getOutputOptions(in char[] sPath, out char[] sXMLOutputFilename, out uint uiNumOfRecsToGen)
{
    toyXML oXML   = new toyXML;
    char[] sCount = ""c;

    //example: sXMLConfig = r"C:\dmd\randomdata\randomdata.xml";
    oXML.loadXMLDocument(sPath ~ r"\randomdata.xml"c);

    sXMLOutputFilename = oXML.getNodeValue("/config/options/output_xml_file"c);
    sCount = oXML.getNodeValue("/config/options/num_of_records_to_generate"c); 
    uiNumOfRecsToGen = std.conv.toUint((sCount == ""c ? "0"c : sCount));

    delete oXML;

} // end void getOutputOptions( in char[], out char[], out uint )

void genFemaleName
( 
    in  udtXmlInfo[] arrNames, 
    in  toyXML[] oXmlNames, 
    out char[] sFirstName, 
    out char[] sMiddleName, 
    out char[] sLastName 
)
{
    char[] sNode;
    uint   uiPos = 0;

    uiPos = getRandomNumber(1, arrNames[FEMALEFIRST].iNumOfRows);
    sNode = arrNames[FEMALEFIRST].sRootTag ~ "/name"c;
    sFirstName = oXmlNames[FEMALEFIRST].getNodeValueByAttr(sNode, "pos"c, uiPos, arrNames[FEMALEFIRST].uiMarkers[uiPos / 100U]);
    //sFirstName = oXmlNames[FEMALEFIRST].getNodeValueByAttr(sNode, "pos"c, uiPos);
        
    uiPos = getRandomNumber(1, arrNames[FEMALEMIDDLE].iNumOfRows);
    sNode = arrNames[FEMALEMIDDLE].sRootTag ~ "/name"c;
    sMiddleName = oXmlNames[FEMALEMIDDLE].getNodeValueByAttr(sNode, "pos"c, uiPos);
        
    uiPos = getRandomNumber(1, arrNames[LASTNAMES].iNumOfRows);
    sNode = arrNames[LASTNAMES].sRootTag ~ "/name"c;
    sLastName = oXmlNames[LASTNAMES].getNodeValueByAttr(sNode, "pos"c, uiPos, arrNames[LASTNAMES].uiMarkers[uiPos / 100U]);
    //sLastName = oXmlNames[LASTNAMES].getNodeValueByAttr(sNode, "pos"c, uiPos);
        
    //writefln( "%s %s %s", sFirstName, sMiddleName, sLastName );
    
} // end genFemaleName( in udtXmlInfo[], in toyXML[], out char[], out char[], out char[] )

void genMaleName
( 
    in  udtXmlInfo[] arrNames, 
    in  toyXML[] oXmlNames, 
    out char[] sFirstName, 
    out char[] sMiddleName, 
    out char[] sLastName 
)
{
    char[] sNode;
    uint uiPos = 0U;

    uiPos = getRandomNumber(1, arrNames[MALEFIRST].iNumOfRows);
    sNode = arrNames[MALEFIRST].sRootTag ~ "/name"c;
    sFirstName = oXmlNames[MALEFIRST].getNodeValueByAttr(sNode, "pos"c, uiPos, arrNames[MALEFIRST].uiMarkers[uiPos / 100U]);
    //sFirstName = oXmlNames[MALEFIRST].getNodeValueByAttr(sNode, "pos"c, uiPos);
        
    uiPos = getRandomNumber(1, arrNames[MALEMIDDLE].iNumOfRows);
    sNode = arrNames[MALEMIDDLE].sRootTag ~ "/name"c;
    sMiddleName = oXmlNames[MALEMIDDLE].getNodeValueByAttr(sNode, "pos"c, uiPos);
        
    uiPos = getRandomNumber(1, arrNames[LASTNAMES].iNumOfRows);
    sNode = arrNames[LASTNAMES].sRootTag ~ "/name"c;
    sLastName = oXmlNames[LASTNAMES].getNodeValueByAttr(sNode, "pos"c, uiPos, arrNames[LASTNAMES].uiMarkers[uiPos / 100U]);
    //sLastName = oXmlNames[LASTNAMES].getNodeValueByAttr(sNode, "pos"c, uiPos);
        
    //writefln( "%s %s %s", sFirstName, sMiddleName, sLastName );
    
} // end genMaleName( in udtXmlInfo[], in toyXML[], out char[], out char[], out char[] )

void genAddress
(
    in  udtXmlInfo[] arrAddrs, 
    in  toyXML[] oXmlAddrs,
    out char[] sAddrNum,
    out char[] sStreetName,
    out char[] sStreetAbbr
)
{
    uint uiNodePos = 0U;
    uint uiPos     = 0U;

    sAddrNum = std.string.toString(getRandomNumber(1, 9999));
 
    uiPos = getRandomNumber(1, arrAddrs[STREETNAME].iNumOfRows);
    sStreetName = oXmlAddrs[STREETNAME].getNodeValueByAttr(arrAddrs[STREETNAME].sRootTag ~ "/street_name"c, "pos"c, uiPos);

    uiPos = getRandomNumber(1, arrAddrs[SUFFIX].iNumOfRows);
    oXmlAddrs[SUFFIX].getNodeValueByAttr(arrAddrs[SUFFIX].sRootTag ~ "/entry"c, "pos"c, uiPos, uiNodePos);
    sStreetAbbr = oXmlAddrs[SUFFIX].getNodeValue("std"c, uiNodePos);

} // end void genAddress( in udtXmlInfo[], in toyXML[], out char[], out char[], out char[] )

void genCityStateZipCode
(
    in  udtXmlInfo[] arrStates,
    in  toyXML[] oXmlStates,
    out char[] sCity, 
    out char[] sState,
    out char[] sZipCode,
    out char[] sAreaCode
)
{
    uint uiNum     = 0U;
    uint uiNodePos = 0U;
    uint uiPos     = 0U;

    uiNum = getRandomNumber(0, arrStates.length - 1);
    uiPos = getRandomNumber(1, arrStates[uiNum].iNumOfRows);
    oXmlStates[uiNum].getNodeValueByAttr(arrStates[uiNum].sRootTag ~ "/entry"c, "pos"c, uiPos, uiNodePos);
 
    sState    = arrStates[uiNum].sXMLFile[0 .. 2];
    sZipCode  = oXmlStates[uiNum].getNodeValue("zipcode"c, uiNodePos);
    sCity     = oXmlStates[uiNum].getNodeValue("city"c, uiNodePos);
    sAreaCode = oXmlStates[uiNum].getNodeValue("areacode"c, uiNodePos);
    
} // end void genCityStateZipCode( in udtXmlInfo[], in toyXML[], out char[], out char[], out char[], out char[] )

// Pass in the age of the person and create a birth date for them.
char[] genBirthDate(in uint uiAge)
{
    static uint[] uiDays = [ 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 ];
    char[] sDate;
    uint   uiDay;
    uint   uiMonth;
    uint   uiYear;
    
    d_time dt;
    
    dt      = getUTCtime();
    uiYear  = YearFromTime(dt) - uiAge;
    uiMonth = getRandomNumber(0, 11); 
    
    if (std.date.LeapYear(uiYear) && uiMonth == 1) 
        uiDay = getRandomNumber(1, 29);
    else
        uiDay = getRandomNumber(1, uiDays[uiMonth]); 
     
    sDate = format("%02s/%02s/%04s", uiMonth + 1, uiDay, uiYear);   
    if (calcAge(sDate) > uiAge)
        uiYear++;
    else if (calcAge(sDate) < uiAge)    
        uiYear--;
        
    if (uiMonth == 1 && uiDay == 29)
        uiDay--;
      
    sDate = format("%02s/%02s/%04s", uiMonth + 1, uiDay, uiYear);  
    
    return sDate;

} // end char[] genBirthDate( in uint )

// < ~ s ~ pos=" ~ toString(ui) ~ "
// example: <name pos="233"
bool setPosTagMarkers(in int iItem, in char[] s, in udtXmlInfo[] arrInfo, in toyXML[] oXML)
{
    char[] sXML        = oXML[iItem].getXML(); 
    uint   uiNumOfRows = arrInfo[iItem].iNumOfRows;
    uint   uiMarkers   = uiNumOfRows / 100U;
    uint   uiNodePos   = 0U;
    uint   uix         = 0U;
    char[] sNodeTag    = "<"c ~ s ~ " pos=\""c;  
    arrInfo[iItem].uiMarkers.length = uiMarkers + 1;

    for (uint ui = 1U; ui <= uiNumOfRows; ui += 100U) 
    {
        uiNodePos = stringplus.ifindx(sXML, (sNodeTag ~ std.string.toString(ui) ~ "\""c), uiNodePos);
        arrInfo[iItem].uiMarkers[uix] = uiNodePos;
        //writef("uiMarkers[%d]=%d ", uix, arrInfo[iItem].uiMarkers[uix]);
        uix++;
    }
    return true;

} // end bool setPosTagMarkers( in int, in char[], in udtXmlInfo[], in toyXML[] )
   
/+
 ' Source        : uint getRandomNumber( in uint = 0x0000, in uint = 0xFFFFFFFF )
 ' Author        : David L. 'SpottedTiger' Davis
 ' Date Created  : 12.Oct.04 Compiled and Tested with dmd v0.102
 ' Modified Date : 28.Feb.05 Compiled and Tested with dmd v0.114
 '               : 26.Jul.06 Fixed an error that appeared if getRandomNumber() 
 '               :           was called using both the default values...I only
 '               :           tested this function with ranges. Compiled and
 '               :           tested with dmd v0.163.  
 ' Requires      : std.random and std.date
 ' -----------------------------------------------------------------------------
 ' By default, returns a unsigned number between 0 to 4,294,967,295. 
 +/
uint getRandomNumber( in uint uiLowNum = 0x0, in uint uiHighNum = 0xFFFFFFFF )
{
    d_time sysTime;
    uint   uiRandom;
    static uint uiMixNum; 
    
    if ( uiHighNum <= 0 )
        return uiLowNum;
        
    if ( uiLowNum == uiHighNum || uiLowNum > uiHighNum )
        return uiLowNum;
    
    sysTime = msFromTime( getUTCtime() );
    uiMixNum += sysTime << 5;
    
    rand_seed( uiMixNum, ( sysTime * SecFromTime( sysTime ) ) - uiMixNum );
    
    if (uiLowNum == 0x0 && uiHighNum == 0xFFFFFFFF)
        uiRandom = ( rand() % uiHighNum );
    else
        uiRandom = ( rand() % ( uiHighNum - uiLowNum + 1 ) + uiLowNum );
    
    //writefln( "uiLowNum=%06d, uiRandom=%06d, uiHighNum=%06d", 
    //           uiLowNum, uiRandom, uiHighNum );

    return uiRandom;
    
} // end uint getRandomNumber( in uint = 0x0, in uint = 0xFFFFFFFF )

int main(in char[][] args)
{
    toyXML[] oXmlStates;
    toyXML[] oXmlNames;
    toyXML[] oXmlAddrs;
    udtXmlInfo[] arrStates;
    udtXmlInfo[] arrNames;
    udtXmlInfo[] arrAddrs;
    // --------------------
    char[] sPath        = ""c;
    uint   uiNum        = 0U;
    uint   uiNodePos    = 0U;
    uint   uiPos        = 0U;
    uint   uiPrevNum    = 0U;
    bool   bGender      = FEMALE;
    char[] sPrevNum     = ""c;
    char[] sAddrNum     = ""c;
    char[] sStreetName  = ""c;
    char[] sStreetAbbr  = ""c;
    char[] sCity        = ""c;
    char[] sState       = ""c;
    char[] sZipCode     = ""c;
    char[] sAreaCode    = ""c;
    uint   uiPhone1     = 0U;
    uint   uiPhone2     = 0U;
    char[] sPhone1      = ""c;
    char[] sPhone2      = ""c;
    // ---------------------
    char[] sFirstName   = ""c;
    char[] sMiddleName  = ""c;
    char[] sLastName    = ""c;
    char[] sSex         = ""c;
    bool   bDisplayWork = true;
    uint   uiCreate     = 0;    //10000;
    char[] sBirthDate   = ""c;
    // ---------------------
    char[] sXMLOutFile  = ""c;   
    StringBuffer sBuffy;

    /+
     ' Set sPath to the Drive and Directory path of the executeable that is running.
     ' (example: sPath = r"C:\dmd\randomempdata")
     ' Also note that the XML files should be in a directory that is called "\xml" above sPath.
     ' (example: r"C:\dmd\randomempdata\xml" the xml is appended in the program.)
     +/
    sPath = args[0][0 .. std.string.rfind(args[0], r"\"c )].dup;

    getOutputOptions(sPath, sXMLOutFile, uiCreate);

    if (sXMLOutFile == ""c || uiCreate == 0U)
        return 0;

    for (uint uix = 1; uix < args[].length; uix++)
    {
        switch (std.string.tolower(args[uix]))
        {
          // Disable displaying random data output to the console.
          case "-nd"c: { bDisplayWork = false; break; }
        }
    }
    
    writefln( "Random personnel names and addresses generator v0.2"c );
    writefln( "by David L. 'SpottedTiger' Davis - 2004-2007"c );
    writefln();

    // Setup and loading XML Files.
    loadStateXMLs(sPath, arrStates, oXmlStates);

    loadNameXMLs(sPath, arrNames, oXmlNames);
    setPosTagMarkers(FEMALEFIRST, "name"c, arrNames, oXmlNames);
    setPosTagMarkers(MALEFIRST, "name"c, arrNames, oXmlNames);
    setPosTagMarkers(LASTNAMES, "name"c, arrNames, oXmlNames);

    loadAddrXMLs(sPath, arrAddrs, oXmlAddrs);
    
    sBuffy.setsize(1024 * 1024 * 5); //5MB
    sBuffy.append("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"c);
    sBuffy.append("<root count=\""c ~ std.string.toString(uiCreate) ~ "\">\n"c);

    for (uint uix = 1; uix <= uiCreate; uix++)
    {
        if (bGender == FEMALE)
        {
            // Generate a female name
            genFemaleName(arrNames, oXmlNames, sFirstName, sMiddleName, sLastName);   
            sSex = "F"c;  
        }
        else
        {
            // Generate a male name
            genMaleName(arrNames, oXmlNames, sFirstName, sMiddleName, sLastName); 
            sSex = "M"c;
        }

        bGender = (bGender == MALE ? FEMALE : MALE);

        // Generate an address
        genAddress(arrAddrs, oXmlAddrs, sAddrNum, sStreetName, sStreetAbbr); 
        genCityStateZipCode(arrStates, oXmlStates, sCity, sState, sZipCode, sAreaCode);

        uiPhone1 = getRandomNumber(1, 999);
        sPhone1 = std.string.zfill(std.string.toString(uiPhone1), 3);

        uiPhone2 = getRandomNumber(1, 9999);
        sPhone2 = std.string.zfill(std.string.toString(uiPhone2), 4);

        sBirthDate = genBirthDate(getRandomNumber(16, 65)); // select ages from 16 to 65

        sBuffy.append("\t<entry pos=\""c, std.string.toString(uix), "\">\n"c,
                      "\t\t<firstname>"c, sFirstName, "</firstname>\n"c,
                      "\t\t<middlename>"c, sMiddleName, "</middlename>\n"c,
                      "\t\t<lastname>"c, sLastName, "</lastname>\n"c,
                      "\t\t<birthdate>"c, sBirthDate, "</birthdate>\n"c,
                      "\t\t<gender>"c, sSex, "</gender>\n"c);
                     
        sBuffy.append("\t\t<streetnum>"c, sAddrNum, "</streetnum>\n"c,
                      "\t\t<streetname>"c, sStreetName, "</streetname>\n"c,
                      "\t\t<streetabbr>"c, sStreetAbbr, "</streetabbr>\n"c,
                      "\t\t<city>"c, sCity, "</city>\n"c,
                      "\t\t<state>"c, sState, "</state>\n"c,
                      "\t\t<zipcode>"c, sZipCode, "</zipcode>\n"c,
                      "\t\t<homephonenum1>("c, sAreaCode, ") "c, sPhone1, "-"c, sPhone2, "</homephonenum1>\n"c,
                      "\t</entry>\n"c);
        
        if (bDisplayWork == true)
        {
            writefln( "%s %s %s"c, sFirstName, sMiddleName, sLastName );
            writefln( "%s %s %s"c, sAddrNum, sStreetName, sStreetAbbr );
            writefln( "%s, %s %s"c, sCity, sState, sZipCode );
            writefln( "(%s) %s-%s"c, sAreaCode, sPhone1, sPhone2 );
            writefln( "Age:%s Birthday:%s "c, calcAge(sBirthDate), sBirthDate );
            writefln();
        }
    } 

    sBuffy.append("</root>"c);

    writefln( "Created %d records."c, uiCreate );
    std.file.write(sPath ~ r"\"c ~ sXMLOutFile, sBuffy.getBuffer);

    writefln( "Wrote \"%s\" to \"%s\""c, sXMLOutFile, sPath );
    writefln( "done."c );

    return 0;

} // end main( in char[][] )