/+
 ' Module          : stringbuffer.d v0.12 
 '                 : a general purpose String buffer to append data into.
 ' Author          : David L. 'SpottedTiger' Davis
 ' Source Platform : Intel x86 / WinXP SP2 
 ' Target Platform : Should work on any CPU / OS combo that D compiles on.
 '                 :
 ' Created Date    : 15.Jul.05 Compiled and tested with dmd v0.128.
 ' Modified Date   : 27.Jul.05 Added the ability to append one StringBuffer's
 '                 :           value into another one.
 '                 : 13.Aug.05 Compiled and tested with dmd v0.129.
 '                 : 06.Sep.05 Added wchar, dchar, wchar[], void[], dchar[],
 '                 :           ifloat, idouble, ireal, cfloat, cdouble, as 
 '                 :           additional datatyes that can be converted and 
 '                 :           appended to the StringBuffer. Plus compiled
 '                 :           and tested with dmd v0.130.
 '                 : 18.Sep.05 Compiled and tested with dmd v0.131.     
 '                 : 26.Oct.05 Recompiled and tested with dmd v0.137. 
 '                 : 30.Dec.05 Recompiled and tested with dmd v0.142.  
 '                 : 22.Mar.06 Recompiled and tested with dmd v0.150. 
 '                 : 28.May.06 Recompiled and Tested with dmd v0.159.  
 '                 : 19.Jul.06 All "Imports" are now private by default 
 '                 :           (which shouldn't be a problem, since I
 '                 :           always use the private keyword) in 
 '                 :           dmd v0.163, recompiled and tested.
 '                 : 23.Oct.06 Recompiled and Tested with dmd v0.172. 
 '                 : 05.Nov.06 Recompiled and Tested with dmd v0.173. 
 '                 : 09.Dec.06 Recompiled and Tested with dmd v0.177. 
 '                 : 05.Jan.07 Recompiled and Tested with dmd v1.0. 
 '                 :
 ' Requirements    : std.stdarg, std.string, and std.utf
 ' Licence         : Public Domain
 ' ---------------------------------------------------------------------------
 ' 
 ' To complie for a unittest:
 ' dmd stringbuffer.d -debug=stringbuffer -unittest
 '
 ' To compile a D program that uses stringbuffer.d:
 ' dmd MySource.d stringbuffer.d
 +/
module stringbuffer;

debug( stringbuffer ) private import std.stdio;
debug( stringbuffer ) private import std.date;
private import std.string;
private import std.stdarg; // for argument gathering
private import std.utf;

struct StringBuffer
{
    private char[] sBuffer;
    private int    iPos  = 0;
    private int    iSize = 0;
    private bool   bAllowOverflow = true;
    private bool   bInit = false;

    void setsize(in int iSetSize = 64) { sBuffer.length = iSetSize; iSize = iSetSize; bInit = true; }
    int  getsize() { return iSize; }
    int  getpos()  { return iPos;  }
    void allowoverflow(in bool b) { bAllowOverflow = b; }
    void clear(in char c = 0x20) { for(int i = 0; i < iSize; i++) sBuffer[i] = c; iPos = 0; }
    //void setmask(in char[] sm = "")
   
    // Main workhorse function.
    private void appendData(in char[] s) 
    {
        int iLen = s.length;
    
        if (!bInit && bAllowOverflow)
            setsize();

        //writefln("iPos=%d, iLen=%d, iBLen=%d", iPos, s.length, sBuffer.length);
        if ((iLen + iPos) <= iSize)
        {
            sBuffer[iPos..iPos + iLen] = s;
            iPos += iLen;
        }
        else if (bAllowOverflow)
        {
            setsize(iPos + iLen);
            sBuffer[iPos..iPos + iLen] = s;   
            iPos += iLen;       
        }
        else if (iSize >= iPos)
        {
            sBuffer[iPos..iSize] = s[0..iSize - iPos];
            iPos = iSize;
        } 
        //else do nothing, parameter is ignored
    }

    void append(...) { append(_arguments, _argptr); }

    void append(in TypeInfo[] _arguments, in va_list _argptr)
    {
        for (int i = 0; i < _arguments.length; i++)
        {   
            //_arguments[i].print();
        
            if (_arguments[i] is typeid(real))
                 appendData(std.string.toString(va_arg!(real)(_argptr)));
             else if (_arguments[i] is typeid(char)) 
                 appendData(std.string.toString(va_arg!(char)(_argptr)));
             else if (_arguments[i] is typeid(wchar))
             {
                 wchar[] ws;
                 ws.length = 1;
                 ws[0..1]  = va_arg!(wchar)(_argptr);
                 appendData(std.utf.toUTF8(ws));
             }
             else if (_arguments[i] is typeid(dchar))
             {
                 dchar[] ds;
                 ds.length = 1;
                 ds[0..1]  = va_arg!(dchar)(_argptr);
                 appendData(std.utf.toUTF8(ds));
             }
             else if (_arguments[i] is typeid(byte))  
                 appendData(std.string.toString(va_arg!(byte)(_argptr)));
             else if (_arguments[i] is typeid(ubyte)) 
                 appendData(std.string.toString( va_arg!(ubyte)(_argptr)));
             else if (_arguments[i] is typeid(short)) 
                 appendData(std.string.toString(va_arg!(short)(_argptr)));
             else if (_arguments[i] is typeid(ushort)) 
                 appendData(std.string.toString(va_arg!(ushort)(_argptr)));
             else if (_arguments[i] is typeid(int))   
                 appendData(std.string.toString(va_arg!(int)(_argptr)));
             else if (_arguments[i] is typeid(uint)) 
                 appendData(std.string.toString(va_arg!(uint)(_argptr)));
             else if (_arguments[i] is typeid(long))   
                 appendData(std.string.toString(va_arg!(long)(_argptr)));
             else if (_arguments[i] is typeid(ulong)) 
                appendData(std.string.toString(va_arg!(ulong)(_argptr)));
             else if (_arguments[i] is typeid(float)) 
                 appendData(std.string.toString(va_arg!(float)(_argptr)));
             else if (_arguments[i] is typeid(double))
                 appendData(std.string.toString(va_arg!(double)(_argptr)));
             else if (_arguments[i] is typeid(char[]))
                 appendData(va_arg!(char[])(_argptr)); 
            else if (_arguments[i] is typeid(wchar[]))
                appendData(std.utf.toUTF8(va_arg!(wchar[])(_argptr)));
            else if (_arguments[i] is typeid(dchar[]))
                appendData(std.utf.toUTF8(va_arg!(dchar[])(_argptr)));
             else if (_arguments[i] is typeid(void[]))
                 appendData(cast(char[])va_arg!(void[])(_argptr)); 
             else if (_arguments[i] is typeid(StringBuffer))
             {
                 StringBuffer sWillow;
                 sWillow.setsize(5);
                 sWillow = va_arg!(StringBuffer)(_argptr); 
                 
                 appendData(sWillow.getBuffer());
             }
            else if (_arguments[i] is typeid(ifloat))
                appendData(std.string.toString(va_arg!(ifloat)(_argptr)));
            else if (_arguments[i] is typeid(idouble))
                appendData(std.string.toString(va_arg!(idouble)(_argptr)));
            else if (_arguments[i] is typeid(ireal))
                appendData(std.string.toString(va_arg!(ireal)(_argptr)));
            else if (_arguments[i] is typeid(cfloat))
                appendData(std.string.toString(va_arg!(cfloat)(_argptr)));
            else if (_arguments[i] is typeid(cdouble))
                appendData(std.string.toString(va_arg!(cdouble)(_argptr)));
            else if (_arguments[i] is typeid(creal))
                appendData(std.string.toString(va_arg!(creal)(_argptr)));

             //else do nothing, parameter is ignored
        }
    }

    char[] getBuffer() { return sBuffer[0..iPos]; }
}

unittest
{
    StringBuffer sBuffy;
    StringBuffer sWillow;
    
    debug( stringbuffer )
    {
        std.date.Date dp;
        d_time dtDate;
        int iMonth = 0;
        int iWeekday = 0;

        dp.parse("Thu, 07/04/1776 12:55:06 UTC"c);
        dtDate = std.date.parse("07/04/1776 12:55:06 UTC"c);

        with (dp)
        {
            sBuffy.setsize(29);
            sBuffy.allowoverflow(false);
            iMonth = (month - 1) * 3;
            iWeekday = (weekday - 1) * 3; 

            sBuffy.append(daystr[iWeekday..iWeekday + 3], ", "c,
                          zfill(std.string.toString(day), 2), " "c,
                          monstr[iMonth..iMonth + 3], " "c,
                          zfill(std.string.toString(year), 3), " "c,
                          zfill(std.string.toString(hour), 2), ":"c,
                          zfill(std.string.toString(minute), 2), ":"c,
                          zfill(std.string.toString(second), 2), 
                          " UTC");
        } 

        assert(sBuffy.getBuffer() == toUTCString(dtDate));
        //writefln("\"%s\" \"%s\"", sBuffy.getBuffer(), toUTCString(dtDate)); 
    }

    sBuffy.setsize(9);
    sBuffy.clear();
    sBuffy.allowoverflow(false);
    sBuffy.append("01/01/1970"c);
    // With the overflow set to false, and the buffer set
    // to allow only 9 characters...the 0 is dropped below.
    assert(sBuffy.getBuffer() == "01/01/197"c);

    sBuffy.setsize(15);
    sBuffy.clear();
    sBuffy.allowoverflow(false);
    sBuffy.append(12, "/", 31, "/", "2010"c);
    assert(sBuffy.getBuffer() == "12/31/2010"c);

    sBuffy.append(cast(char)0x20, "UTC"c);
    assert(sBuffy.getBuffer() == "12/31/2010 UTC"c);

    sBuffy.setsize(5);
    sBuffy.clear();
    sBuffy.allowoverflow(true);
    sBuffy.append("0123456789"c);
    sBuffy.append("-"c);
    sBuffy.append("ABC"c);
    assert(sBuffy.getBuffer() == "0123456789-ABC"c);
    //writefln("\"%s\" size=%d", sBuffy.getBuffer(), sBuffy.getsize());
    assert(sBuffy.getsize() == 14);
    
    // Testing appending one StringBuffer into another
    sWillow.append("Willow"c);
    sBuffy.setsize(1);
    sBuffy.clear();
    sBuffy.allowoverflow(true);
    sBuffy.append(sWillow);
    sBuffy.append(" is Buffy's best friend."c);
    //writefln("sWillow=\"%s\", sBuffy=\"%s\" size=%d", 
    //          sWillow.getBuffer(), sBuffy.getBuffer(), sBuffy.getsize());
    assert(sWillow.getBuffer() == "Willow"c);
    assert(sBuffy.getBuffer() == "Willow is Buffy's best friend."c);
    assert(sBuffy.getsize() == 30);
}

debug( stringbuffer )
{
int main()
{
    std.stdio.writefln( "stringbuffer.unittest finished!"c );
    return 0;
    
} // end int main()    
}