/+
 ' Program         : toyxml.d v0.24 - toyXML Class (case insensitive)
 ' Author          : David L. 'SpottedTiger' Davis
 ' Source Platform : Intel x86 / WinXP SP2 
 ' Target Platform : Should work on any CPU / OS combo that D compiles on.
 '                 :
 ' Complier        : DigitalMars "D" aka Mars v0.101
 ' Created Date    : 01.Sep.04 Compiled and tested with dmd v0.101.
 ' Modified Date   : 16.Nov.04 Compiled and tested with dmd v0.106,
 '                 :           and added a unittest section.
 '                 : 25.Nov.04 Made some corrections in both the
 '                 :           getChildNodes() and getNodeValue() functions.
 '                 : 30.Nov.04 Made a few more corrections, and tighted up
 '                 :           the code a little bit more, then
 '                 :           compiled and tested with dmd v0.108.
 '                 :           Also improved both getNumOfNodes() and 
 '                 :           getNodeValueByPosition() functions to work  
 '                 :           better with multiple-mixed NodeTags, like 
 '                 :           <NodeTag> and <NodeTag Attr="x"> to gather
 '                 :           a correct count or position.
 '                 : 29.Jan.05 Recompiled and tested with dmd v0.111.
 '                 : 10.Jun.05 Recompiled and tested with dmd v0.126.
 '                 : 17.Sep.05 Recompiled and tested with dmd v0.131.
 '                 : 18.Sep.05 Added the getXML() function.
 '                 : 26.Oct.05 Recompiled and Tested with dmd v0.137.
 '                 : 13.Jan.06 Recompiled and Tested with dmd v0.143.
 '                 : 26.Mar.06 Recompiled and Tested with dmd v0.150.
 '                 : 28.May.06 Recompiled and Tested with dmd v0.159.
 '                 : 19.Jul.06 All "Imports" are now private by default  
 '                 :           (which shouldn't be a problem, since I  
 '                 :           always use the private keyword) in 
 '                 :           dmd v0.163, recompiled and tested.
 '                 : 05.Nov.06 Recompiled and unittested under D v0.173.
 '                 : 25.Nov.06 Compiled and tested with dmd v0.175.
 '                 : 09.Dec.06 Compiled and tested with dmd v0.177.
 '                 : 05.Jan.07 Compiled and tested with dmd v1.0.
 '                 :
 ' Requirements    : std.string, std.file, and stringplus.d v0.13
 '                 : ( std.stdio for debugging and unittests )
 '                 :
 ' Licence         : Public Domain
 ' ---------------------------------------------------------------------------
 ' Note: This class only works will XML in plain text, and currently its 
 '       methods only read data from its content.
 '
 ' Public functions:
 ' -----------------
 ' 01) void loadXMLDocument( in char[], bool = true )
 ' 02) ulong ulXMLDocumentLength()
 ' 03) void saveXMLDocument( in char[] )
 ' 04) char[] getChildNodes( in char[], in uint = 0U )
 ' 05) uint getNumOfNodes( in char[] )
 ' 06) char[] getNodeValueByPosition( in char[], in uint )
 ' 07) char[] getAttrValue( in char[], in char[] ) 
 ' 08) char[] getNodeValueByAttr( in char[], in char[], in char[] )
 ' 09) char[] getNodeValueByAttr( in char[], in char[], in char[], out uint )
 ' 10) char[] getNodeValueByAttr( in char[], in char[], in uint )
 ' 11) char[] getNodeValueByAttr( in char[], in char[], in int, out uint )
 ' 12) char[] getNodeValue( in char[], in uint = 0U )
 ' 13) char[] getXML(in uint = 0U, in uint = uint.max )
 '
 ' Private functions:
 ' ------------------
 ' 01) char[] loadXMLFile( in char[] )
 ' 02) char[] doInitialCheck( in char[] )
 '
 ' =====================================
 ' To compile for a unittest:
 ' dmd toyxml.d stringplus.d -debug=toyxml -unittest 
 '
 ' To compile into a project:
 ' dmd myproject.d toyxml.d stringplus.d
 '
 +/
module toyxml;

debug( toyxml ) private import std.stdio;  // for writefln()
//private import std.stdio; 
private import std.string; // for strip(), split(), and rfind()
private import stringplus; // for ifindx(), and irfindx()
private import std.file;   // for getSize(), read() and write()

/+
 ' toyXML class (-WIP-)
 ' ------------------------------
 ' This class was build to load the complete content of the 
 ' XML document in to memory as a large string. And its methods 
 ' currently, only read data from its contents.
 +/
public class toyXML 
{
    private char[] sXMLDOM = ""c;
    
    /+
     ' void loadXMLDocument( in char[] )
     '
     ' To load from a memory buffer: oXML.loadXMLDocument( sXmlBuf );
     ' To load from a harddisk file: oXML.loadXMLDocument( r"C:\dmd\two_sets_of_names.xml" );
     +/
    public void loadXMLDocument(in char[] sXmlBuffer, in bool bRemoveTabs = true)
    {
        if (sXmlBuffer.length == 0)
        {
            sXMLDOM = ""c;    
            return ;
        }
        
        if (sXmlBuffer[0] != '<')
            sXMLDOM = loadXMLFile(sXmlBuffer);
        else
        {
            sXMLDOM.length = sXmlBuffer.length;
            sXMLDOM = sXmlBuffer.dup; 
        }
        
        if (bRemoveTabs == true)
            sXMLDOM = std.string.replace(sXMLDOM, "\t"c, ""c);
        
    } // end void loadXMLDocument( in char[] )
   
    public char[] getXML(in uint uiBegPos = 0U, in uint uiEndPos = uint.max) 
    { 
        if (uiEndPos < uiBegPos || uiEndPos > sXMLDOM.length)
            uiEndPos = sXMLDOM.length;

        return sXMLDOM[uiBegPos .. uiEndPos]; 
    } 

    public ulong ulXMLDocumentLength() { return sXMLDOM.length; } 
    
    private char[] loadXMLFile(in char[] sFile)
    {
        sXMLDOM.length = std.file.getSize(sFile);
        return cast(char[])std.file.read(sFile);
        
    } // end char[] loadXML( in char[] )
   
    public void saveXMLDocument(in char[] sFile)
    {
        if (sFile.length == 0)
            return;
    
        std.file.write(sFile, sXMLDOM);
   
    } // end void saveXMLDocument( in char[] )

    /+
     ' char[] getNodeValue( in char[], in uint = 0U )
     '
     ' example: Gather "xxxxx" value between <usa_names><last_names><name>xxxxx</name></last_names></usa_names>
     ' sNodeBuf = oXML.getNodeValue( "/usa_names/last_names/name" )
     '            -OR-
     ' <pa_zipcodes><entry Zipcode="15764"><Zipcode>99999</ZipCode></entry></pa_zipcodes>
     ' sNodeValue = oXML.getNodeValueByAttr( "/pa_zipcodes/entry", "Zipcode", 15764, iNodePos );
     ' sNodeValue = oXML.getNodeValue( "ZipCode", iNodePos );  
     '
     ' Returns the Child Node's value.
     +/       
    public char[] getNodeValue(in char[] sNodePrm, in uint uiPos = 0U)
    {
        char[] sBuf        = ""c;
        char[] sNode       = sNodePrm.dup;
        char[] sNodeParent = ""c;
        char[] sNodeChild  = ""c;
        int    iBgnXsdPos  = 0;
        int    iEndXsdPos  = 0;
        int    iSlashPos   = 0; 
        char   cChar;

        sNode = doInitialCheck(sNode);
        
        if (sNode.length == 0)
            return ""c;

        if (std.string.count( sNode, "/"c ) > 1)
        {
            iSlashPos = std.string.rfind(sNode, "/"c);
            sNodeParent = sNode[0 .. iSlashPos];
            sBuf = getChildNodes(sNodeParent, uiPos);
            
            if (sBuf.length <= 0)
                return ""c;
            
            sNodeChild = sNode[iSlashPos + 1 .. sNode.length];
            iBgnXsdPos = stringplus.ifindx(sBuf, "<"c ~ sNodeChild, uiPos); 
            
            // Checks that the Node ( "<NodeTagName" has a space or a ">" following the Nodetag name.
            if ( iBgnXsdPos != -1 )
            {
                cChar = sBuf[iBgnXsdPos + sNodeChild.length + 1];
                if ( cChar != '\x20' && cChar != '>') 
                    return ""c;
                else if (cChar == '\x20')
                    iBgnXsdPos = stringplus.ifindx(sBuf, ">"c, iBgnXsdPos + sNodeChild.length + 1); 
                else
                    iBgnXsdPos = iBgnXsdPos + sNodeChild.length + 1;      
            }
           
            if (iBgnXsdPos >= 0)
            {
                iEndXsdPos = stringplus.ifindx(sBuf, "</"c ~ sNodeChild ~ ">"c, iBgnXsdPos); 

                if (iEndXsdPos > iBgnXsdPos)
                    return sBuf[(iBgnXsdPos + 1) .. iEndXsdPos].dup;
                else
                    return ""c;    
            }
        }
        else
        {
            // Replaces the "/" for a single node
            sNode = sNode[1 .. sNode.length];
            iBgnXsdPos = stringplus.ifindx(sXMLDOM,  "<"c ~ sNode, uiPos); 
 
            // Checks that the Node ( "<NodeTagName" has a space or a ">" following the Nodetag name.
            if (iBgnXsdPos != -1)
            {
                cChar = sXMLDOM[iBgnXsdPos + sNode.length + 1];
                if (cChar != '\x20' && cChar != '>') 
                    return ""c;
            }
   
            if (iBgnXsdPos >= 0)
            {
                iEndXsdPos = stringplus.ifindx(sXMLDOM, "</"c ~ sNode ~ ">"c, iBgnXsdPos); 

                if (iEndXsdPos > iBgnXsdPos)
                    return sXMLDOM[(iBgnXsdPos + sNode.length + 2) .. iEndXsdPos].dup;
                else
                    return ""c;    
            }
        }

        return ""c;

    } // end char[] getNodeValue( in char[], in uint = 0U )

    /+
     ' char[] getChildNodes( in char[], in uint = 0U )
     '
     ' example: Gather all nodes between <usa_names><last_names>...<usa_names></last_names>
     ' sNodeBuf = oXML.getChildNodes( "/usa_names/last_names" )
     '            -OR-
     ' oXML.getNodeValueByAttr( "/pa_zipcodes/entry", "Zipcode", 15764, uiDataPos, iNodePos );
     ' sNodeValue = oXML.getNodeValue( "ZipCode", iNodePos );
     '
     ' Returns all the Childern Nodes for the selected Parent Node.
     +/
    public char[] getChildNodes(in char[] sNodePar, in int uiPos = 0U) 
    {
        char[][] sarrNodeLbl;   // array of strings
        char[]   sBuf           = ""c;
        int      iNodeBgnPos    = uiPos; //0;
        int      iNodeEndPos    = sXMLDOM.length;
        int      iNodeAdjBgnPos = -1;
        int      iNodeAdjEndPos = -1;
        int      iNode          = 0U;
        int      iLastNode      = 0U;
        char[]   sNode          = sNodePar.dup;
        char     cChar;
        
        sNode = doInitialCheck(sNode);
        
        if (sNode.length == 0)
            return ""c;
                
        if (sNode[0U] == '/') 
            sNode = sNode[1U .. sNode.length];
        
        if (sNode[sNode.length - 1U .. sNode.length] == "/"c)
            sNode = sNode[0U .. sNode.length - 1U];
    
        sarrNodeLbl = std.string.split(sNode, "/"c); 
        if (uiPos > 0U)
            iLastNode = sarrNodeLbl.length - 1;
    
        if (sarrNodeLbl.length > 1)
        {
            for (iNode = iLastNode; iNode < sarrNodeLbl.length; iNode++)
            {
                sNode = r"<"c ~ sarrNodeLbl[iNode][];
                
                // Find partial node label
                iNodeAdjBgnPos = stringplus.ifindx(sXMLDOM, sNode, iNodeBgnPos);
// -- Testing --
/+
                // Checks that the Node ( "<NodeTagName" has a space or a ">" following the Nodetag name.
                if ( iNodeAdjBgnPos != -1 )
                {
                    cChar = sXMLDOM[ iNodeAdjBgnPos + sNode.length + 1 ];
                    //writefln( "cChar=%s", cChar );
                    if ( cChar != '\x20' && cChar != '>' ) 
                        return "";
                }
+/
// -------------                
                if (iNodeAdjBgnPos > iNodeBgnPos) 
                    iNodeBgnPos = iNodeAdjBgnPos + sNode.length;
                
                // Find the end of the partial node label
                iNodeAdjBgnPos = stringplus.ifindx(sXMLDOM, ">"c, iNodeBgnPos);
                
                if (iNodeAdjBgnPos > iNodeBgnPos) 
                    iNodeBgnPos = iNodeAdjBgnPos + 1;
                
                sNode = r"</"c ~ sarrNodeLbl[iNode][] ~ ">"c;
                iNodeAdjEndPos = stringplus.irfindx(sXMLDOM, sNode, iNodeEndPos);
                
                if (iNodeAdjEndPos != -1 && iNodeAdjEndPos < iNodeEndPos) 
                    iNodeEndPos = iNodeAdjEndPos;
                
                if (sXMLDOM[iNodeBgnPos .. iNodeEndPos].sizeof < 512)
                {
                    iNodeAdjEndPos = stringplus.ifindx(sXMLDOM, sNode, iNodeBgnPos);
                    
                    if (iNodeAdjEndPos != -1 && iNodeAdjEndPos < iNodeEndPos) 
                        iNodeEndPos = iNodeAdjEndPos;
                }
            }
        }    
    
        if (iNodeBgnPos < 0 || iNodeEndPos > sXMLDOM.length || iNodeBgnPos == iNodeEndPos)   
            return ""c;

        if (sXMLDOM[iNodeBgnPos] == '>')
            sBuf = std.string.strip(sXMLDOM[iNodeBgnPos + 1 .. iNodeEndPos]);
        else
            // Handles the very first node which won't have a ">" before it.
            sBuf = std.string.strip(sXMLDOM[iNodeBgnPos .. iNodeEndPos]);
        
        if (std.string.find(sBuf, "<"c) != -1)
            return sBuf.dup;
        else
            return ""c;    
        
    } // end char[] getChildNodes( in char[], in uint = 0U )

   /+
    ' uint getNumOfNodes( in char[] )
    '
    ' example: <usa_names><last_names1><name> gets the count of <name>s found.
    ' lNumOfNodes = oXML.getNumOfNodes( "/usa_names/last_names1/name" )
    '
    ' Returns the number of nodes that have the same 
    ' node tag-name within a group of nodes, or zero.
    +/
    public uint getNumOfNodes(in char[] sNodePrm)
    {
        int    iLblBgnPos[3] = 0;
        //--------------------------
        int    iLblEndPos  = 0;
        char[] sNodeLbl1   = ""c;
        char[] sNodeLbl2   = ""c;
        char[] sXmlBuf     = ""c;
        char[] sNode       = sNodePrm.dup;
        uint   uiCounter   = 0U;
        
        sNode = doInitialCheck(sNode);
        
        if (sNode.length == 0)
            return 0U;
        
        if (std.string.count(sNode, "/"c ) > 1) 
        {
            // A child node.
            iLblEndPos = std.string.rfind(sNode, "/"c);
            sNodeLbl1 = "<"c ~ sNode[iLblEndPos + 1 .. sNode.length] ~ "\x20"c;
            sNodeLbl2 = "<"c ~ sNode[iLblEndPos + 1 .. sNode.length] ~ ">"c;
            sNode  = sNode[0 .. iLblEndPos];
        }    
        else
        {    
            // The root node.
            sNodeLbl1 = "<"c ~ sNode[1 .. sNode.length] ~ "\x20"c;
            sNodeLbl2 = "<"c ~ sNode[1 .. sNode.length] ~ ">"c;
        }
        
        sXmlBuf = getChildNodes(sNode);
        
        if (sXmlBuf.length < 1)
            return 0U;
            
        iLblEndPos = sXmlBuf.length;
        
        while ( true )
        {
            // Seek the 1st version of the Node that has an Attr.
            iLblBgnPos[1] = stringplus.ifindx(sXmlBuf, sNodeLbl1, iLblBgnPos[0]);
            iLblBgnPos[2] = stringplus.ifindx(sXmlBuf, sNodeLbl2, iLblBgnPos[0]);
            
            if (iLblBgnPos[1] == -1 && iLblBgnPos[2] == -1)
                break;
            else if (iLblBgnPos[1] == -1)   
                iLblBgnPos[0] = iLblBgnPos[2];
            else if (iLblBgnPos[2] == -1) 
                iLblBgnPos[0] = iLblBgnPos[1];  
            else if (iLblBgnPos[1] < iLblBgnPos[2])  
                iLblBgnPos[0] = iLblBgnPos[1];
            else
                iLblBgnPos[0] = iLblBgnPos[2];
                
            // Find the position of the current Node's ending delimiter.    
            iLblEndPos = stringplus.ifindx(sXmlBuf, ">"c, iLblBgnPos[0]); 
 
            // If the current Node's ending delimiter 
            // found, add 1 to the counter.
            if (iLblEndPos > -1) 
                uiCounter++;
            else
                break;       

            // Continue looking, move the position forward
            iLblBgnPos[0] = iLblEndPos;
        }
        
        return uiCounter;

    } // end uint getNumOfNodes( in char[] )    
    
    /+
     ' char[] getNodeValueByPosition( in char[], in uint )
     '
     ' example: <usa_names><last_names2><name>xxxxx</name> gets n number of <name>'s value. 
     ' sNodeValue = oXML.getNodeValueByPosition( "/usa_names/last_names2/name", 9 )
     '
     ' Returns the value held by a given node tag-name 
     ' in a given postion, or an empty string. 
     +/
    public char[] getNodeValueByPosition(in char[] sNodePrm, in uint uiElementPos)
    {
        int    iLblBgnPos[3] = 0;
        //--------------------------
        int    iLblEndPos  = 0;
        char[] sNodeLbl1   = ""c;
        char[] sNodeLbl2   = ""c;
        char[] sXmlBuf     = ""c;
        char[] sNode       = sNodePrm.dup;
        uint   uiCurPos    = 0U;

        sNode = doInitialCheck(sNode);
 
        if (sNode.length == 0)
            return ""c;
       
        if (std.string.count(sNode, "/"c) > 1) 
        {
            // A child node.
            iLblEndPos = std.string.rfind(sNode, "/"c);
            sNodeLbl1 = "<"c ~ sNode[iLblEndPos + 1 .. sNode.length] ~ "\x20"c;
            sNodeLbl2 = "<"c ~ sNode[iLblEndPos + 1 .. sNode.length] ~ ">"c;
            sNode  = sNode[0 .. iLblEndPos];
        }    
        else
        {    
            // The root node.
            sNodeLbl1 = "<"c ~ sNode[1 .. sNode.length] ~ "\x20"c;
            sNodeLbl2 = "<"c ~ sNode[1 .. sNode.length] ~ ">"c;
        }
        
        sXmlBuf = getChildNodes(sNode);
        
        if (sXmlBuf.length < 1)
            return ""c;
            
        iLblEndPos = sXmlBuf.length;
        
        // Walk the Tree
        while (true)
        {
            // Seek the 1st version of the Node that has an Attr.
            iLblBgnPos[1] = stringplus.ifindx(sXmlBuf, sNodeLbl1, iLblBgnPos[0]);
            iLblBgnPos[2] = stringplus.ifindx(sXmlBuf, sNodeLbl2, iLblBgnPos[0]);
            
            if (iLblBgnPos[1] == -1 && iLblBgnPos[2] == -1)
                break;
            else if (iLblBgnPos[1] == -1 )   
                iLblBgnPos[0] = iLblBgnPos[2];
            else if (iLblBgnPos[2] == -1 ) 
                iLblBgnPos[0] = iLblBgnPos[1];  
            else if (iLblBgnPos[1] < iLblBgnPos[2] )  
                iLblBgnPos[0] = iLblBgnPos[1];
            else
                iLblBgnPos[0] = iLblBgnPos[2];
                
            // Find the position of the current Node's ending delimiter.    
            iLblEndPos = stringplus.ifindx(sXmlBuf, ">"c, iLblBgnPos[0]); 
 
            // If the current Node's ending delimiter 
            // found, add 1 to the counter.
            if (iLblEndPos > -1) 
                uiCurPos++;
            else
                break;
 
            // If somehow the current Node's position greater, break out.
            if (uiCurPos > uiElementPos) 
                break;
            
            // If correct Node is found, return the it's value. 
            if (uiCurPos == uiElementPos) 
            {
                iLblBgnPos[0] = stringplus.ifindx(sXmlBuf, ">"c, iLblBgnPos[0]); 
                iLblEndPos = stringplus.ifindx(sXmlBuf, "</"c, iLblBgnPos[0]);
            
                return sXmlBuf[iLblBgnPos[0] + 1 .. iLblEndPos]; 
            }
            
            // Continue looking, move the position forward
            iLblBgnPos[0] = iLblEndPos;
        }
    
        return ""c;
        
    } // end char[] getNodeValueByPosition( in char[], in uint )
    
    /+
     ' char[] getAttrValue( in char[], in char[] ) 
     '
     ' example: <usa_names><last_names1 Count="nnnn"> gets "nnnn"
     ' sAttrValue = oXML.getAttrValue( "/usa_names/last_names1", "Count" );
     '
     ' Returns a Node's Attr value
     +/
    public char[] getAttrValue(in  char[] sNodePrm, in  char[] sAttrPrm)
    {
        int    iNodeNxtPos = 0;
        int    iLblBgnPos  = 0;
        int    iLblEndPos  = 0;
        char[] sNodeLbl1   = ""c;
        char[] sNodeLbl2   = ""c;
        char[] sXmlBuf     = ""c;
        char[] sNode       = sNodePrm.dup;
        char[] sAttr       = sAttrPrm.dup ~ "="c;
        int    iCurPos     = 0;

        sNode = doInitialCheck(sNode);

        if (sNode.length == 0)
            return ""c;
        
        if (std.string.count(sNode, "/") > 1) 
        {
            iLblEndPos = std.string.rfind(sNode, "/"c);
            sNodeLbl1 = "<"c ~ sNode[iLblEndPos + 1 .. sNode.length] ~ "\x20"c;
            sNodeLbl2 = "<"c ~ sNode[iLblEndPos + 1 .. sNode.length] ~ ">"c;
            sNode  = sNode[0 .. iLblEndPos];
        }    
        else
        {    
            sNodeLbl1 = "<"c ~ sNode[1 .. sNode.length] ~ "\x20"c;
            sNodeLbl2 = "<"c ~ sNode[1 .. sNode.length] ~ ">"c;
        }
        
        sXmlBuf = getChildNodes(sNode);
        iLblEndPos = sXmlBuf.length;

        while ( true )
        {
            // Find Child Node
            iNodeNxtPos = stringplus.ifindx(sXmlBuf, sNodeLbl1, iLblBgnPos);
            
            if (iNodeNxtPos != -1)
                iLblBgnPos = iNodeNxtPos;
            else
                return ""c;    
            
            // Find ending delimiter of Child Node's 
            iLblEndPos = stringplus.ifindx(sXmlBuf, ">"c, iLblBgnPos);
            
            if (iLblEndPos == -1) 
                return ""c;
            
            // Find AttrTag or AttrTag with value, bit it must be within the Child Node
            iNodeNxtPos = stringplus.ifindx(sXmlBuf, sAttr, iLblBgnPos);
            
            if ( iNodeNxtPos == -1 ) 
                return ""c;
            else if (iNodeNxtPos > iLblEndPos) 
            {
                // Found AttrTag, but it's in another Node, retry.
                iLblBgnPos = stringplus.irfindx(sXmlBuf, "<"c, iNodeNxtPos);
                
                if (iNodeNxtPos == -1) 
                    return ""c;

                continue;
            } 
                        
            break;
        }    
 
        // Find 1st quote      
        iNodeNxtPos = stringplus.ifindx(sXmlBuf, "\""c, iNodeNxtPos + 1);

        if (iNodeNxtPos != -1)
        {
            iLblBgnPos = iNodeNxtPos + 1;
                
            // Find 2nd quote
            iNodeNxtPos = stringplus.ifindx(sXmlBuf, "\""c, iLblBgnPos + 1);
            
            if (iNodeNxtPos != -1)
                iLblEndPos = iNodeNxtPos;
            else
                return ""c;
        }
        else
            return ""c;
 
        return sXmlBuf[iLblBgnPos .. iLblEndPos].dup;  
          
    } // end char[] getAttrValue( in char[], in char[] )    

    public char[] getNodeValueByAttr
    ( 
        in  char[] sNodePrm, 
        in  char[] sAttrPrm, 
        in  int    iXmlAttrValue
    )
    {
        uint uiNodePos;
    
        return getNodeValueByAttr(sNodePrm, sAttrPrm, std.string.toString(iXmlAttrValue), uiNodePos);

    } // end char[] getNodeValueByAttr( in char[], in char[], in int )
        
    /+
     ' char[] getNodeValueByAttr( in char[], in char[], in int, out uint )
     '
     ' Example: <usa_names><last_names2><name Pos="6">xxxxx</name> gets "xxxxx"
     ' sNodeValue = oXML.getNodeValueByAttr( "/usa_names/last_names2/name", "Pos", 6 );
     '
     ' Returns the Node's value based on the Attr's name and value.
     ' Also returns the position within the XML text in the "out" parameter.
     +/
    public char[] getNodeValueByAttr
    ( 
        in  char[] sNodePrm, 
        in  char[] sAttrPrm, 
        in  int    iXmlAttrValue,
        out uint   uiNodePos 
    )
    {
        return getNodeValueByAttr(sNodePrm, sAttrPrm, std.string.toString(iXmlAttrValue), uiNodePos);
    
    } // end char[] getNodeValueByAttr( in char[], in char[], in int, out uint )
    
    public char[] getNodeValueByAttr
    ( 
        in  char[] sNodePrm, 
        in  char[] sAttrPrm, 
        in  char[] sAttrValue
    )
    {
        uint uiNodePos;
    
        return getNodeValueByAttr(sNodePrm, sAttrPrm, sAttrValue, uiNodePos);

    } // end char[] getNodeValueByAttr( in char[], in char[], in char[] )
        
    /+
     ' char[] getNodeValueByAttr( in char[], in char[], in char[], out uint )
     '
     ' Example: <usa_names><last_names2><name Pos="6">xxxxx</name> gets "xxxxx"
     ' sNodeValue = oXML.getNodeValueByAttr( "/usa_names/last_names2/name", "Pos", "6" );
     '
     ' Returns the Node's value based on the Attr's name and value.
     +/
    public char[] getNodeValueByAttr
    ( 
        in  char[] sNodePrm, 
        in  char[] sAttrPrm, 
        in  char[] sAttrValue,
        out uint   uiNodePos,
        in  uint   uiBgnPos = 0U 
    )
    {
        int    iNodeNxtPos = 0;
        int    iLblBgnPos  = uiBgnPos; // 0;
        int    iLblEndPos  = 0;
        char[] sNode       = sNodePrm.dup;
        char[] sAttr       = sAttrPrm ~ "=\""c ~ sAttrValue ~ "\""c;
        char[] sXmlBuf     = ""c;
        char[] sNodeLbl1   = ""c;
        char[] sTempBuf    = ""c; 
       
        uiNodePos = 0U;
        sNode = doInitialCheck(sNode);
 
        if (sNode.length == 0)
            return ""c;
       
        // Must be the Node above the Root Node
        if (std.string.count(sNode, "/"c) > 1) 
        {
            iLblEndPos = std.string.rfind(sNode, "/"c);
            sNodeLbl1 = "<"c ~ sNode[iLblEndPos + 1 .. sNode.length] ~ "\x20"c;
            sNode  = sNode[0 .. iLblEndPos];
        }    
        else
            // At the Root Node
            sNodeLbl1 = "<"c ~ sNode[1 .. sNode.length] ~ "\x20"c;
        
        // sXmlBuf points to a sub-region of sXMLDOM
        sXmlBuf = getChildNodes(sNode);
        iLblEndPos = sXmlBuf.length;

        while (true)
        {
            // Find Child Node
            iNodeNxtPos = stringplus.ifindx(sXmlBuf, sNodeLbl1, iLblBgnPos);
            
            if (iNodeNxtPos != -1)
                iLblBgnPos = iNodeNxtPos;
            else
                return ""c;    
            
            // Find ending delimiter of Child Node's 
            iLblEndPos = stringplus.ifindx(sXmlBuf, ">"c, iLblBgnPos);
            
            if (iLblEndPos == -1) 
                return ""c;
            
            // Find AttrTag or AttrTag with value, but it must be within the Child Node
            iNodeNxtPos = stringplus.ifindx(sXmlBuf, sAttr, iLblBgnPos);
 
            if (iNodeNxtPos == -1) 
                return ""c;

            else if (iNodeNxtPos > iLblEndPos) 
            {
                // Found AttrTag, but it's in another Node, retry.
                iLblBgnPos = stringplus.irfindx(sXmlBuf, "<"c, iNodeNxtPos);
                
                if (iNodeNxtPos == -1) 
                    return ""c;

                continue;
            }    

            // Now move iLblBgnPos to the ">" position
            iLblBgnPos = iLblEndPos;
    
            // Now set the iLblEndPos to the "</" position 
            iLblEndPos = stringplus.ifindx(sXmlBuf, "</"c, iLblBgnPos);
            
            if (iLblEndPos == -1) 
                return ""c;
        
            break;
        }
        
        // Bookmark the beginning of the current NodeTag
        iNodeNxtPos = stringplus.irfindx(sXmlBuf, sNodeLbl1, iLblEndPos);
        uiNodePos   = (iNodeNxtPos == -1 ? 0U : iNodeNxtPos);
        sTempBuf    = std.string.strip(sXmlBuf[iLblBgnPos + 1 .. iLblEndPos]);
        
        if (sTempBuf[0] != '<')
            return sTempBuf.dup;
        else
            return ""c;    
        
    } // end char[] getNodeValueByAttr( in char[], in char[], in char[], out uint )

    private char[] doInitialCheck(in char[] sNode)
    {
        // If XML Doc hasn't been loaded return "".
        if (sXMLDOM.length == 0U) 
            return ""c;

        // Can't find a "" or null node, return "".
        if (sNode.length < 1)
            return ""c;

        // If the beginning "/" is missing, append to sNode
        if (sNode[0 .. 1] != "/"c)
            sNode = "/"c ~ sNode;
       
        return sNode;

    } // end char[] doInitialCheck( in char[] )

} // end class toyXML

unittest
{
    debug( toyxml ) writefln( "toyxml.unittests running..."c );
    
    toyXML oXMLDoc     = new toyXML;
    toyXML oXMLSubDoc  = new toyXML;
    char[] sXmlBuf     = ""c;
    char[] sAttrValue  = ""c;
    char[] sNodeValue  = ""c;
    long   lNumOfNodes = 0;
    uint   uiNodePos   = 0U;
    
    sXmlBuf = "<usa_names type=\"collection\">\n"c
                "<last_names1 type=\"list\" count=\"10\">\n"c
                  "<name type=\"string\" pos=\"1\">Smith</name>\n"c
                  "<name type=\"string\" pos=\"2\">Johnson</name>\n"c
                  "<name type=\"string\" pos=\"3\">Williams</name>\n"c
                  "<name type=\"string\" pos=\"4\">Jones</name>\n"c
                  "<name type=\"string\" pos=\"5\">Brown</name>\n"c
                  "<name type=\"string\" pos=\"6\">Davis</name>\n"c
                  "<name type=\"string\" pos=\"7\">Miller</name>\n"c
                  "<name type=\"string\" pos=\"8\">Wilson</name>\n"c
                  "<name type=\"string\" pos=\"9\">Moore</name>\n"c
                  "<name type=\"string\" pos=\"10\">Taylor</name>\n"c
                "</last_names1>\n"c
        
                "<last_names2 type=\"list\" count=\"10\">\n"c
                  "<name type=\"string\" pos=\"1\">Anderson</name>\n"c
                  "<name type=\"string\" pos=\"2\">Thomas</name>\n"c
                  "<name type=\"string\" pos=\"3\">Jackson</name>\n"c
                  "<name type=\"string\" pos=\"4\">White</name>\n"c
                  "<name type=\"string\" pos=\"5\">Harris</name>\n"c
                  "<name type=\"string\" pos=\"6\">Martin</name>\n"c
                  "<name type=\"string\" pos=\"7\">Thompson</name>\n"c
                  "<name type=\"string\" pos=\"8\">Garcia</name>\n"c
                  "<name type=\"string\" pos=\"9\">Martinez</name>\n"c
                  "<name type=\"string\" pos=\"10\">Robinson</name>\n"c
                "</last_names2>\n"c
              "</usa_names>"c;

    debug( toyxml ) writefln( "toyxml.loadXMLDocument( in char[] ).unittest"c );
    oXMLDoc.loadXMLDocument(sXmlBuf);
    assert( oXMLDoc.ulXMLDocumentLength() == sXmlBuf.length );
    
    debug( toyxml ) writefln( "toyxml.getChildNodes( in char[] ).unittest"c );              
    sXmlBuf = oXMLDoc.getChildNodes( "/usa_names/last_names1/name"c );
    assert( sXmlBuf == ""c );
    
    sXmlBuf = oXMLDoc.getChildNodes("/usa_names/last_names2"c );
    lNumOfNodes = oXMLDoc.getNumOfNodes( "/usa_names/last_names2/name"c );  
    assert( lNumOfNodes == 10 );
    
    debug( toyxml ) writefln( "toyxml.getAttrValue( in char[], in char[] ).unittest"c );                 
    sAttrValue = oXMLDoc.getAttrValue( "/usa_names/last_names2", "Count"c );
    assert( sAttrValue == "10"c );
    
    debug( toyxml ) writefln( "toyxml.getNodeValueByAttr( in char[], in char[], in char[] ).unittest"c ); 
    sNodeValue = oXMLDoc.getNodeValueByAttr( "/usa_names/last_names2/name", "Pos", "3"c ); 
    assert( sNodeValue == "Jackson"c );
    
    debug( toyxml ) writefln( "toyxml.getNodeValueByPosition( in char[], in uint ).unittest"c ); 
    sNodeValue = oXMLDoc.getNodeValueByPosition( "/usa_names/last_names2/name"c, 9 );
    assert( sNodeValue == "Martinez"c );
    
    debug( toyxml ) writefln( "toyxml.getNodeValueByAttr( in char[], in char[], in uint ).unittest" ); 
    sNodeValue = oXMLDoc.getNodeValueByAttr( "/usa_names/last_names1/name", "Pos", 8 ); 
    assert( sNodeValue == "Wilson"c );
    
    sXmlBuf = "<pa_zipcodes count=\"2832\">\n"c
                "<entry zipcode=\"18068\" pos=\"1580\">\n"c
                  "<zipcode>18068</zipcode>\n"c
                  "<city>Old Zionsville</city>\n"c
                  "<areacode>610</areacode>\n"c
                  "<county>Lehigh</county>\n"c
                "</entry>\n"c
                "<entry zipcode=\"19547\" pos=\"1581\">\n"c
                  "<zipcode>19547</zipcode>\n"c
                  "<city>Oley</city>\n"c
                  "<areacode>610</areacode>\n"c
                  "<county>Berks</county>\n"c
                "</entry>\n"c
                "<entry zipcode=\"15764\" pos=\"1582\">\n"c
                  "<zipcode>15764</zipcode>\n"c
                  "<city>Oliveburg</city>\n"c
                  "<areacode>814</areacode>\n"c
                  "<county>Jefferson</county>\n"c
                "</entry>\n"c
                "<entry zipcode=\"15472\" pos=\"1583\">\n"c
                  "<zipcode>15472</zipcode>\n"c
                  "<city>Oliver</city>\n"c
                  "<areacode>724</areacode>\n"c
                  "<county>Fayette</county>\n"c
                "</entry>\n"c
                "<entry zipcode=\"18448\" pos=\"1584\">\n"c
                  "<zipcode>18448</zipcode>\n"c
                  "<city>Olyphant</city>\n"c
                  "<areacode>717</areacode>\n"c
                  "<county>Lackawanna</county>\n"c
                "</entry>\n"c
              "</pa_zipcodes>"c;
              
    /+
     ' Below the getNodeValueByAttr( "/pa_zipcodes/entry", "Zipcode", 15764, iDataPos, iNodePos )
     ' is being used to find a node using an Attr and its value to pin-point its location within
     ' the XML Doc, so data can be extracted from the child nodes.
     '
     ' example:
     ' ... 
     '  <entry zipcode="15764" pos="1582">
     '    <zipcode>15764</zipcode>
     '    <city>Oliveburg</city>
     '    <areacode>814</areacode>
     '    <county>Jefferson</county>
     '  </entry>
     ' ...
     +/
    //oXMLDoc.loadXMLDocument( r"C:\dmd\PA_ZipCodes.xml" );
    oXMLDoc.loadXMLDocument( sXmlBuf );
    
    sNodeValue = oXMLDoc.getNodeValueByAttr( "/pa_zipcodes/entry"c, "Zipcode"c, 15764, uiNodePos );
    assert( sNodeValue == ""c );
    assert( uiNodePos == 308 );
    
    // This method uses the same XML Doc
    sNodeValue = oXMLDoc.getNodeValue( "ZipCode"c, uiNodePos );
    assert( sNodeValue == "15764"c );
    
    sNodeValue = oXMLDoc.getNodeValue( "City"c, uiNodePos );
    assert( sNodeValue == "Oliveburg"c );
    
    sNodeValue = oXMLDoc.getNodeValue( "AreaCode"c, uiNodePos );
    assert( sNodeValue == "814"c );
    
    sNodeValue = oXMLDoc.getNodeValue( "County"c, uiNodePos ); 
    assert( sNodeValue == "Jefferson"c );  
    
    /+
     ' This method builds a new XML Doc using a SubDoc
     '
     ' example:
     '    <zipcode>15764</zipcode>
     '    <city>Oliveburg</city>
     '    <areacode>814</areacode>
     '    <county>Jefferson</county>     
     +/    
    sXmlBuf = oXMLDoc.getChildNodes("/pa_zipcodes/entry"c, uiNodePos );
 
    oXMLSubDoc.loadXMLDocument( sXmlBuf );
    
    sNodeValue = oXMLSubDoc.getNodeValue( "ZipCode"c );
    assert( sNodeValue == "15764"c );
    
    sNodeValue = oXMLSubDoc.getNodeValue( "City"c );
    assert( sNodeValue == "Oliveburg"c );
    
    sNodeValue = oXMLSubDoc.getNodeValue( "AreaCode"c );
    assert( sNodeValue == "814"c );
    
    sNodeValue = oXMLSubDoc.getNodeValue( "County"c );
    assert( sNodeValue == "Jefferson"c );  

    // -----------------------------------------
    sNodeValue = oXMLDoc.getAttrValue( "pa_zipcodes"c, "Count"c ); 
    assert( sNodeValue == "2832"c );  

    sNodeValue = oXMLDoc.getAttrValue( "/pa_zipcodes"c, "Count"c ); 
    assert( sNodeValue == "2832"c );  

    /+
     ' Load up and test against a real config.xml file, which
     ' also has a unittest node for even better testing.
     +/
    sXmlBuf = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"c
              "<config>\n"c
                "<usa>\n"c
                  "<states_selected>AL,AK,AZ,AR,CA,CO,CT,DE,DC,FL,GA,HI,ID,IL,IN,IA,KS,KY,LA,ME,MD,MA,MI,MN,MS,MO,MT,NE,NV,NH,NJ,NM,NY,NC,ND,OH,OK,OR,PA,PR,RI,SC,SD,TN,TX,UT,VT,VA,WA,WV,WI,WY</states_selected>\n"c
                  "<!-- example: if AL is selected, then the filename becomes AL_ZipCodes.xml -->\n"c
                  "<zipcode_fileext>_ZipCodes.xml</zipcode_fileext>\n"c
                  "<state_list>AL,AK,AZ,AR,CA,CO,CT,DE,DC,FL,GA,HI,ID,IL,IN,IA,KS,KY,LA,ME,MD,MA,MI,MN,MS,MO,MT,NE,NV,NH,NJ,NM,NY,NC,ND,OH,OK,OR,PA,PR,RI,SC,SD,TN,TX,UT,VT,VA,WA,WV,WI,WY</state_list>\n"c
                  "<designators roottag=\"designators\">USA_Sec_Unit_Designator_Abbr.xml</designators>\n"c
                  "<suffixes roottag=\"street_suffixes\">USA_Street_Suffixes_Abbr.xml</suffixes>\n"c
                  "<streets roottag=\"usa_streets\">USA_Street_Names.xml</streets>\n"c
                  "<female_first_names roottag=\"female_first_names\">USA_Female_First_Names.xml</female_first_names>\n"c
                  "<female_middle_names roottag=\"female_middle_names\">USA_Female_Middle_Names.xml</female_middle_names>\n"c
                  "<male_first_names roottag=\"male_first_names\">USA_Male_First_Names.xml</male_first_names>\n"c
                  "<male_middle_names roottag=\"male_middle_names\">USA_Male_Middle_Names.xml</male_middle_names>\n"c
                  "<last_names roottag=\"last_names\">USA_Last_Names.xml</last_names>\n"c
                  "<bothsexes_middle_names roottag=\"bothsexes_middle_names\">USA_BothSexes_Middle_Names.xml</bothsexes_middle_names>\n"c
                "</usa>\n"c
                "<unittest>\n"c
                   "<!-- Testing with NodeTagName, Attr, AttrValue, and NodeValue all have the same string. -->\n"c
                   "<test1 test1=\"test1\">test1</test1>\n"c
                   "<test2 pos=\"1\">test2</test2>\n"c
                   "<test1>test1b</test1>\n"c
                   "<test2>test2b</test2>\n"c
                   "<test1 test1=\"test1\" pos=\"3\">test1c</test1>\n"c
                   "<test10>test10</test10>\n"c
                "</unittest>\n"c
              "</config>"c;
    
    oXMLDoc.loadXMLDocument( sXmlBuf );    
    
    sAttrValue = oXMLDoc.getAttrValue( "/config/usa/designators"c, "roottag"c );
    assert( sAttrValue == "designators"c );
    
    sNodeValue = oXMLDoc.getNodeValue( "/config/unittest/test1"c );
    assert( sNodeValue == "test1"c );
    
    sAttrValue = oXMLDoc.getAttrValue( "/config/unittest/test1"c, "test1"c );
    assert( sAttrValue == "test1"c );
   
    sNodeValue = oXMLDoc.getNodeValueByAttr( "/config/unittest/test1"c, "pos"c, 3, uiNodePos );
    assert( sNodeValue == "test1c"c );
    
    sNodeValue = oXMLDoc.getNodeValueByAttr( "/config/unittest/test1"c, "pos"c, "3"c, uiNodePos );
    assert( sNodeValue == "test1c"c );
    
    sNodeValue = oXMLDoc.getNodeValueByPosition( "/config/unittest/test1"c, 2 );
    assert( sNodeValue == "test1b"c );
    
    lNumOfNodes = oXMLDoc.getNumOfNodes( "/config/unittest/test1"c );
    assert( lNumOfNodes == 3 );

    lNumOfNodes = oXMLDoc.getNumOfNodes( "/config/unittest/test2"c );
    assert( lNumOfNodes == 2 ); 
    
    lNumOfNodes = oXMLDoc.getNumOfNodes( "/config/unittest/test10"c );
    assert( lNumOfNodes == 1 );   
    
    sNodeValue = oXMLDoc.getNodeValue( "/config/unittest/test10"c );
    assert( sNodeValue == "test10"c ); 
    
    sNodeValue = oXMLDoc.getNodeValueByPosition( "/config/unittest/test2"c, 2 );
    assert( sNodeValue == "test2b"c );
    
    sNodeValue = oXMLDoc.getNodeValueByPosition( "/config/unittest/test1"c, 3 );
    assert( sNodeValue == "test1c"c );  
    
    sNodeValue = oXMLDoc.getNodeValueByAttr( "/config/usa/bothsexes_middle_names"c, "roottag"c, "bothsexes_middle_names"c );
    assert( sNodeValue == "USA_BothSexes_Middle_Names.xml"c );    
    
    // Do function tests with a totally empty XML Document, no errors should appear.
    // All values should return as a "" (empty string), and positions / counts as 0 (zero).
    oXMLDoc.loadXMLDocument( ""c );

    sNodeValue = oXMLDoc.getNodeValue( "/root/tree/branch/leaf"c );
    assert( sNodeValue == ""c ); 

    sNodeValue = oXMLDoc.getNodeValueByPosition( "/root/tree/branch/leaf"c, 3 );
    assert( sNodeValue == ""c );  
    
    sNodeValue = oXMLDoc.getNodeValueByAttr( "/root/tree/branch/leaf"c, "pos"c, "6"c, uiNodePos );
    assert( sNodeValue == ""c );
    assert( uiNodePos == 0 );
    
    sNodeValue = oXMLDoc.getNodeValueByAttr( "/root/tree/branch/leaf", "pos", 3, uiNodePos );
    assert( sNodeValue == ""c );
    assert( uiNodePos == 0 );
    
    sAttrValue = oXMLDoc.getAttrValue( "/root/tree/branch/leaf"c, "Color"c );
    assert( sAttrValue == ""c );

    sXmlBuf = oXMLDoc.getChildNodes( "/root/tree/branch/leaf"c );
    assert( sXmlBuf == ""c );  

    lNumOfNodes = oXMLDoc.getNumOfNodes( "/root/tree/branch/leaf"c );  
    assert( lNumOfNodes == 0 );
    
    assert( oXMLDoc.ulXMLDocumentLength() == 0 );
}

debug( toyxml )
{
int main()
{
    writefln( "toyxml.unittests done.\n"c );
    
    return 0;
}
} // end debug